'use strict';

const NodeModule = require('node:module');

function _interopNamespaceCompat(e) {
    if (e && typeof e === 'object' && 'default' in e) return e;
    const n = Object.create(null);
    if (e) {
        for (const k in e) {
            if (k !== 'default') {
                const d = Object.getOwnPropertyDescriptor(e, k);
                Object.defineProperty(n, k, d.get ? d : {
                    enumerable: true,
                    get: function () { return e[k]; }
                });
            }
        }
    }
    n.default = e;
    return Object.freeze(n);
}

const NodeModule__namespace = /*#__PURE__*/_interopNamespaceCompat(NodeModule);

/**
 * Checks whether given data is a boolean
 * @param {any} aData The data to examine
 * @return {boolean} The result
 */
function isBoolean(aData) {
    return typeof aData === 'boolean';
}

/**
 * Checks whether given data is an object
 * @param {any} aData The data to examine
 * @return {boolean} The result
 */
function isObject(aData) {
    return typeof aData === 'object' && aData !== null && !Array.isArray(aData);
}

/**
 * Checks whether given data is a node-module
 * @param {any} aData The data to examine
 * @return {boolean} The result
 */
function isNodeModule(aData) {
    return aData instanceof NodeModule__namespace.Module;
}

/**
 * A simple helper class, that acts as data cell
 */
class Cell {
    /**
     * The constructor
     */
    constructor() {
        this._data = null;
    }

    /**
     * Returns the content of the cell
     * @return {any} The cell data
     */
    get() {
        return this._data;
    }

    /**
     * Writes given data to the cell
     * @param {any} aNewData The data to write
     */
    set(aNewData) {
        this._data = aNewData;
    }
}

/**
 * Defines a module confinement
 */
class ModuleConfinement {
    /**
     * The constructor for a ModuleConfinement
     * @param {Object} aRawConfinement
     * @param {NodeModule} aModuleContext
     */
    constructor(aRawConfinement, aModuleContext) {
        const moduleFileNameHelper = (aModuleName) => NodeModule__namespace._resolveFilename(aModuleName, aModuleContext, false);
        const rawConfinement = typeof aRawConfinement === 'object' && aRawConfinement !== null ? aRawConfinement : {};

        /**
         * @type {boolean}
         */
        this.allowBuiltIns = isBoolean(rawConfinement.allowBuiltIns) ? rawConfinement.allowBuiltIns : true;
        /**
         * @type {boolean}
         */
        this.applyToChildren = isBoolean(rawConfinement.allowBuiltIns) ? rawConfinement.applyToChildren : false;

        /**
         * @type {Set<string>}
         */
        this.whiteList = new Set();
        /**
         * @type {Set<string>}
         */
        this.blackList = new Set();
        /**
         * @type {Map<string, string>}
         */
        this.redirect = new Map();

        if (Array.isArray(rawConfinement.whiteList) && rawConfinement.whiteList.length > 0) {
            for (let i = 0, iLen = rawConfinement.whiteList.length; i < iLen; i++) {
                const modulePath = moduleFileNameHelper(rawConfinement.whiteList[i]);

                this.whiteList.add(modulePath);
            }
        }

        if (Array.isArray(rawConfinement.blackList) && rawConfinement.blackList.length > 0) {
            for (let i = 0, iLen = rawConfinement.blackList.length; i < iLen; i++) {
                const modulePath = moduleFileNameHelper(rawConfinement.blackList[i]);

                this.blackList.add(modulePath);
            }
        }

        if (isObject(rawConfinement.redirect)) {
            const keys = Object.keys(rawConfinement.redirect);

            for (let i = 0, iLen = keys.length; i < iLen; i++) {
                const redirectFrom = moduleFileNameHelper(keys[i]);
                const redirectTo = moduleFileNameHelper(rawConfinement.redirect[keys[i]]);

                this.redirect.set(redirectFrom, redirectTo);
            }
        }
    }
}

/**
 * The root module
 * @type {Cell<NodeModule>}
 */
const rootModuleCell = new Cell();

/**
 * The default confinement
 * @type {Cell<ModuleConfinement>}
 */
const defaultConfinementCell = new Cell();

/**
 * The specific module confintments
 * @type {Map<string, ModuleConfinement>}
 */
const confinementsMap = new Map();

/**
 * This function decides, whether given module is allowed to get called with given confinement
 * @param {ModuleConfinement} aConfinement
 * @param {string} aModule
 * @return {boolean}
 */
function isAllowedToCall(aConfinement, aModule) {
    const notBlackListed = !aConfinement.blackList.has(aModule);
    const isBuiltInModule = NodeModule__namespace.builtinModules.includes(aModule);
    const allowBuiltIns = aConfinement.allowBuiltIns;

    return aConfinement.whiteList.has(aModule) ||
        !isBuiltInModule && notBlackListed ||
        allowBuiltIns && isBuiltInModule && notBlackListed;
}

/**
 * Tries to find the confinement for given module
 * @param {NodeModule} aModuleContext
 * @param {string} aModule
 * @return {ModuleConfinement | null}
 */
function findConfinementFor(aModuleContext) {
    const rootModule = rootModuleCell.get();
    let targetConfinement = defaultConfinementCell.get();

    if (confinementsMap.has(aModuleContext.id)) {
        targetConfinement = confinementsMap.get(aModuleContext.id);
    }
    else {
        for (let it = aModuleContext.parent; isNodeModule(it) && it !== rootModule; it = it.parent) {
            if (confinementsMap.has(it.id)) {
                const confinement = confinementsMap.get(it.id);

                if (confinement.applyToChildren) {
                    targetConfinement = confinement;
                    break;
                }
            }
        }
    }

    return targetConfinement;
}

/**
 * Installs a proxy for the require function in node, confining all modules based on shared rules
 */
function installRequireProxy() {
    NodeModule__namespace.Module.prototype.require = new Proxy(NodeModule__namespace.Module.prototype.require, {
        apply(aTarget, aThisContext, aArgumentsList) {
            const moduleFileName = NodeModule__namespace._resolveFilename(aArgumentsList[0], aThisContext, false);
            const confinementToFulfill = findConfinementFor(aThisContext);

            if (confinementToFulfill !== null && !isAllowedToCall(confinementToFulfill, moduleFileName)) {
                throw new Error(`NodeModuleConfinement: Confinement violation found! Loading "${moduleFileName}" from "${aThisContext.id}" is not allowed!`);
            }

            const argumentList = aArgumentsList.slice();
            if (confinementToFulfill.redirect.has(moduleFileName)) {
                argumentList[0] = confinementToFulfill.redirect.get(moduleFileName);
            }

            return Reflect.apply(aTarget, aThisContext, argumentList);
        },
    });
}

/**
 * Installs a trap for global.Function, to prevent usage of that
 */
function installTrapFunctionAddon() {
    global.Function = new Proxy(global.Function, {
        construct() {
            throw new Error('NodeModuleConfinement: Found call of "new Function()", which is not allowed.');
        },
        apply() {
            throw new Error('NodeModuleConfinement: Found call of "Function()", which is not allowed.');
        },
    });
}

/**
 * Installs a trap for global.eval, to prevent usage of that
 */
function installTrapEvalAddon() {
    global.eval = new Proxy(global.eval, {
        apply() {
            throw new Error('NodeModuleConfinement: Found call of "eval()", which is not allowed.');
        },
    });
}

/**
 * Installs all traps referenced by the options
 * @param {AddonsConfig} aOptions
 */
function installAddons(aOptions) {
    if (aOptions.trapEval === true) {
        installTrapEvalAddon();
    }

    if (aOptions.trapFunction === true) {
        installTrapFunctionAddon();
    }
}

/**
 * @typedef RawModuleConfinement
 * @property {boolean} [allowBuiltIns=true]
 * @property {boolean} [applyToChildren=false]
 * @property {Array<string>} [whiteList=[]]
 * @property {Array<string>} [blackList=[]]
 * @property {Object} [redirect={}]
 */

/**
 * @typedef AddonsConfig
 * @property {boolean} [trapEval=false]
 * @property {boolean} [trapFunction=false]
 */

/**
 * @typedef NodeModuleConfinementOptions
 * @property {RawModuleConfinement} [defaultConfinement={}]
 * @property {Object} [confinements={}]
 * @property {AddonsConfig} [addons={}]
 */

/**
 * Configures all confinements
 * @param {NodeModule} aRootModule
 * @param {NodeModuleConfinementOptions} aOptions
 */
function setup(aRootModule, aOptions) {
    // first validate the given arguments
    if (!isNodeModule(aRootModule)) {
        throw new TypeError(`NodeModuleConfinement configure: First parameter needs to be an instance of NodeModule`);
    }
    else if (!isObject(aOptions)) {
        throw new TypeError(`NodeModuleConfinement configure: Second parameter needs to be an object, got ${typeof aOptions}`);
    }

    // then write all option data to the shared references
    rootModuleCell.set(aRootModule);

    if (isObject(aOptions.defaultConfinement)) {
        defaultConfinementCell.set(new ModuleConfinement(aOptions.defaultConfinement, aRootModule));
    }
    else {
        defaultConfinementCell.set(new ModuleConfinement({}, aRootModule));
    }

    if (isObject(aOptions.confinements)) {
        const keys = Object.keys(aOptions.confinements);

        for (let i = 0, iLen = keys.length; i < iLen; i++) {
            const targetModule = NodeModule__namespace._resolveFilename(keys[i], aRootModule, false);
            const confinement = new ModuleConfinement(aOptions.confinements[keys[i]], aRootModule);

            confinementsMap.set(targetModule, confinement);
        }
    }

    // if any traps should get installed, call the trap installer
    if (isObject(aOptions.addons)) {
        installAddons(aOptions.addons);
    }

    // Then install the require proxy, that actually prevents bad things from happening
    installRequireProxy();
}

module.exports.setup = setup;
